@@ -66,6 +66,11 @@ class CloudFlare(object):
             else:
                 self.logger = None
 
+        def __del__(self):
+            if self.network:
+                del self.network
+                self.network = None
+
         def call_with_no_auth(self, method, parts,
                               identifier1=None, identifier2=None, identifier3=None, identifier4=None,
                               params=None, data=None, files=None):
@@ -857,17 +862,21 @@ class CloudFlare(object):
                 else:
                     branch = getattr(branch, element)
             except:
-                # should never happen
-                raise CloudFlareAPIError(0, 'api load name failed')
-        name = a[-1]
+                # missing path - should never happen unless api_v4 is a busted file
+                branch = None
+                break
 
+        if not branch:
+                raise CloudFlareAPIError(0, 'api load: element **%s** missing when adding path /%s' % (element, '/'.join(a)))
+
+        name = a[-1]
         try:
             if '-' in name:
                 f = getattr(branch, name.replace('-','_'))
             else:
                 f = getattr(branch, name)
-            # already exists - don't let it overwrite
-            raise CloudFlareAPIError(0, 'api duplicate name found: %s/**%s**' % ('/'.join(a[0:-1]), name))
+            # already exists - don't let it overwrite - should never happen unless api_v4 is a busted file
+            raise CloudFlareAPIError(0, 'api load: duplicate name found: %s/**%s**' % ('/'.join(a[0:-1]), name))
         except AttributeError:
             # this is the required behavior - i.e. it's a new node to create
             pass
@@ -913,6 +922,10 @@ class CloudFlare(object):
                 # it's a known api call - lets show the result and continue down the tree
                 if 'delete' in d or 'get' in d or 'patch' in d or 'post' in d or 'put' in d:
                     # only show the result if a call exists for this part
+                    if '_parts_unused' in d:
+                            # This is an uncallable endpoint - presently no way to return this info
+                            # w.append(str(a)[1:-1] + ' ; UNUSED')
+                            pass
                     if '_parts' in d:
                         if n[-1] == '_' and keyword.iskeyword(n[:-1]):
                             # remove the extra keyword postfix'ed with underscore
@@ -964,9 +977,19 @@ class CloudFlare(object):
         self._base = self._v4base(config)
 
         # add the API calls
-        api_v4(self)
-        if 'extras' in config and config['extras']:
-            api_extras(self, config['extras'])
+        try:
+            api_v4(self)
+            if 'extras' in config and config['extras']:
+                api_extras(self, config['extras'])
+        except Exception as e:
+            raise CloudFlareAPIError(0, str(e))
+
+    def __del__(self):
+        """ Network for Cloudflare API"""
+
+        if self._base:
+            del self._base
+            self._base = None
 
     def __call__(self):
         """ Cloudflare v4 API"""
